from __future__ import annotations

import abc
import asyncio
import dataclasses as dc
import logging
import typing as t
from collections import defaultdict
from contextvars import ContextVar

from . import utils
from .html import HTML
from .messages import Command
from .messages import Event
from .server import Server
from .sessions import Session

logger = logging.getLogger(__name__)


@dc.dataclass
class UpdateLayout(Command):
    html: HTML
    vars: dict = dc.field(default_factory=dict)


@dc.dataclass
class LayoutUpdated(Event): ...


@dc.dataclass
class SetVar(Command):
    name: str
    value: str


@dc.dataclass
class VarSet(Event): ...


@dc.dataclass
class ButtonClick(Event):
    button_id: str


@dc.dataclass
class InputUpdated(Event):
    name: str
    value: str


type Callback = t.Callable[[Event], t.Awaitable[None]]
type CallbacksMap = list[tuple[Event.T, Callback]]

type _CpName = str  # component name
type _CbName = str  # callback name
type _ComponentRegistry = dict[_CpName, list[tuple[Event.T, _CbName]]]

_registry: _ComponentRegistry = defaultdict(list)


def on(event_cls: Event.T) -> t.Callable:
    def wrapper(callback: Callback) -> Callback:
        self = utils.get_f_self(callback)
        cmp_cls_name = utils.get_f_cls_name(callback)

        if self is None and cmp_cls_name:
            _registry[cmp_cls_name].append((event_cls, callback.__name__))
        else:
            raise RuntimeError

        return callback
    return wrapper


class Component:
    html: HTML

    @dc.dataclass
    class Vars: pass

    def __init__(self):
        self.vars = self.Vars()

    @property
    def callbacks_map(self) -> CallbacksMap:
        return [
            (event_cls, getattr(self, callback_name))
            for event_cls, callback_name
            in _registry[self.__class__.__name__]
        ]

    async def update_var(self, name: str) -> None:
        value = getattr(self.vars, name)
        await Session.get().send_command(SetVar(name=name, value=value))


class HTMLComponent(Component):
    """Proxy component for plain HTML string"""
    def __init__(self, html: HTML):
        super().__init__()
        self.html = html


class Page(list[Component]):

    @property
    def callbacks_map(self) -> CallbacksMap:
        cb_map = []
        for component in self:
            cb_map += component.callbacks_map
        return cb_map


type RawPage = t.Iterable[type[Component] | HTML]
type Route = str


def convert_to_page(raw_page: RawPage) -> Page:
    page = Page()
    for item in raw_page:
        if isinstance(item, str):
            component = HTMLComponent(html=item)
        else:
            component = item()

        page.append(component)
    return page


_layout: ContextVar[Layout] = ContextVar('_app_layout')


class Layout:
    def __init__(self, raw_page: RawPage | None = None):
        self.pages: dict[Route, Page] = {}
        self.current_route = None

        if raw_page is not None:
            self.add_page(route=self.current_route, page=raw_page)

    def add_page(self, route: Route, page: RawPage) -> None:
        if not self.current_route:
            self.current_route = route

        self.pages[route] = convert_to_page(page)

    def switch_page(self, route: Route) -> None:
        self.current_route = route

    @property
    def current_page(self) -> Page:
        return self.pages[self.current_route]

    @property
    def html(self) -> HTML:
        return ''.join((comp.html for comp in self.current_page))

    @property
    def vars(self) -> dict:
        result = {}
        for comp in self.current_page:
            result.update(comp.vars.__dict__)
        return result

    @property
    def callbacks_map(self) -> CallbacksMap:
        return self.current_page.callbacks_map

    async def send_update(self) -> None:
        session = Session.get()
        command = UpdateLayout(html=self.html, vars=self.vars)
        await session.send_command(command)


class AppMixinInterface(abc.ABC):
    @abc.abstractmethod
    async def on_session_open(self) -> None: ...

    @abc.abstractmethod
    async def on_session_close(self) -> None: ...

    @abc.abstractmethod
    async def on_event(self, event: Event) -> None: ...


class App(AppMixinInterface):
    def __init__(self):
        self.raw_pages: dict[Route, RawPage] = {}

    @property
    def session(self) -> Session:
        return Session.get()

    @property
    def layout(self) -> Layout:
        return _layout.get()

    @layout.setter
    def layout(self, value: Layout | None) -> None:
        _layout.set(value)

    async def on_session_open(self) -> None:
        layout = Layout()
        for route, page in self.raw_pages.items():
            layout.add_page(route, page)

        self.layout = layout
        await self.layout.send_update()

    async def on_session_close(self) -> None:
        self.layout = None

    async def on_event(self, event: Event) -> None:
        for event_cls, callback in self.layout.callbacks_map:
            if event_cls == event._cls:
                await callback(event)

    async def switch_page(self, route: Route):
        if route not in self.raw_pages:
            raise ValueError(f'Incorrect route: `{route}`')

        self.layout.switch_page(route)
        await self.layout.send_update()

    async def run(
        self,
        page: RawPage = None,
        *,
        routed_pages: dict[Route, RawPage] = None,
    ) -> None:
        assert page or routed_pages
        if page is not None:
            self.raw_pages = {'*': page}

        elif routed_pages is not None:
            self.raw_pages = routed_pages

        self.server = Server(
            on_session_open=self.on_session_open,
            on_session_close=self.on_session_close,
            on_event=self.on_event,
        )
        await self.server.run()

    def run_sync(
        self,
        page: RawPage = None,
        *,
        routed_pages: dict[Route, RawPage] = None,
    ) -> None:
        try:
            asyncio.run(self.run(page, routed_pages=routed_pages))
        except KeyboardInterrupt:
            pass



from __future__ import annotations

import abc
import dataclasses as dc


class _Message(abc.ABC):
    type Name = str
    type T = type[_Message]

    _cls = property(lambda self: self.__class__)
    _name = property(lambda self: self._cls.__name__)
    _data = property(lambda self: dc.asdict(self))


@dc.dataclass
class Event(_Message):
    """`Client` -> `Server` message
    """
    type T = type[Event]

    @classmethod
    def get_by_name(cls, name: Event.Name) -> Event.T:
        return {ec.__name__: ec for ec in Event.__subclasses__()}[name]


@dc.dataclass
class Command(_Message):
    """`Server` -> `Client` message
    """
    type T = type[Command]



# FIXME
#
#  This module is bad idea. No need to work with HTML from Python, it's not OK
#
#  * For simple cases, HTML can be passed as string to `Component`
#  * For more complex cases, HTML should be set in Jinja-template or frontend

import typing as t
from functools import partial

type HTML = str

type Tag = HTML
type TagItems = t.Iterable[HTML]


BUTTON: Tag = '<button {params}>{inner}</button>'
TABLE: Tag = '<table {params}>{items}</table>'
TR: Tag = '<tr {params}>{items}</tr>'
TH: Tag = '<th {params}>{inner}</th>'
TD: Tag = '<td {params}>{inner}</td>'


def _html_params(params: dict) -> HTML:
    return ''.join('{k}={v} ' for k, v in params.items())


def container_tag(tag: Tag, items: TagItems, **params) -> HTML:
    return tag.format(
        items=''.join(items),
        params=_html_params(params),
    )


def text_tag(tag: Tag, inner: str, **params) -> HTML:
    return tag.format(
        inner=inner,
        params=_html_params(params),
    )


button = partial(text_tag, BUTTON)

table = partial(container_tag, TABLE)
tr = partial(container_tag, TR)
th = partial(text_tag, TH)
td = partial(text_tag, TD)



import asyncio
import logging
from contextvars import ContextVar

from .app import App
from .app import AppMixinInterface
from .messages import Event
from .sessions import Session

logger = logging.getLogger(__name__)


class EverySecond(Event): ...


_scheduler_task: ContextVar[asyncio.Task] = ContextVar('_scheduler_task')


class SchedulerMixin(AppMixinInterface):

    async def on_session_open(self) -> None:
        await super().on_session_open()
        task = asyncio.create_task(self.scheduler())
        _scheduler_task.set(task)

    async def on_session_close(self) -> None:
        await super().on_session_close()
        _scheduler_task.get().cancel()
        _scheduler_task.set(None)

    async def scheduler(self):
        event = EverySecond()
        session = Session.get()
        try:
            with session:
                while True:
                    await self.on_event(event)
                    await asyncio.sleep(1)

        except asyncio.CancelledError:
            pass
        except Exception as e:
            logger.exception(e)


class SchedulerApp(SchedulerMixin, App): ...



from .app import Component
from .html import HTML
from .html import table
from .html import td
from .html import th
from .html import tr

type _DatasheetRow = tuple[str, ...]
type _Datasheet = tuple[_DatasheetRow, ...]  # first row using as data header


def render_table(data: _Datasheet) -> HTML:
    headers, *rows = data

    html_headers = tr(th(item) for item in headers)
    html_items = [tr(td(item) for item in row) for row in rows]

    return table([html_headers, *html_items])


# Framework adapter

class TableComponent(Component):
    table_data: _Datasheet = None

    def __init__(self):
        super().__init__()
        if self.table_data is None:
            raise ValueError('`table_data` param is missing')

        self.html = render_table(self.table_data)



import asyncio
import logging
import os
import typing as t

import uvicorn
from starlette.responses import HTMLResponse
from starlette.staticfiles import StaticFiles
from starlette.types import Receive
from starlette.types import Scope
from starlette.types import Send
from starlette.websockets import WebSocket

from .logging import log_config
from .messages import Event
from .sessions import Session
from .sessions import AbstractSession
from .sessions import SessionClosed

logger = logging.getLogger(__name__)

type _ServerCallback[T] = t.Callable[[T], t.Awaitable[None]]

EXIT_CODE = 1

response_404 = HTMLResponse(content='<b>404</b> Not found', status_code=404)
response_405 = HTMLResponse(content='<b>405</b> Not allowed', status_code=405)


class _ASGIServer(uvicorn.Server):
    # Shutdown default OS signals catch
    # It should be implemented on another layer
    def install_signal_handlers(self) -> None: ...


class Server:
    ALLOWED_STATIC_FILES = ('.html', 'css', '.js', '.map', '.ico')

    STATIC_DIR = os.path.dirname(__file__) + '/web'
    INDEX_HTML_PATH = STATIC_DIR + '/index.html'

    def __init__(
        self,
        *,
        on_session_open: _ServerCallback[None],
        on_session_close: _ServerCallback[None],
        on_event: _ServerCallback[Event],
        host: str = 'localhost',
        port: int = 5000,
        session_cls: type[AbstractSession] = Session,
    ):
        self.host, self.port = host, port
        self.static_files = StaticFiles(directory=self.STATIC_DIR)

        self.on_session_open = on_session_open
        self.on_session_close = on_session_close
        self.on_event = on_event
        self.session_cls = session_cls

    async def run(self) -> None:
        config = uvicorn.Config(
            app=self,
            host=self.host,
            port=self.port,
            log_level='debug',
            log_config=log_config,
        )
        server = _ASGIServer(config=config)

        logger.info('Starting server')
        try:
            await server.serve()
        finally:
            logger.info('Shutting down server')

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        if scope['type'] == 'lifespan':
            exit_code = await self.handle_lifespan(scope, receive, send)
            if exit_code:
                return

        elif scope['type'] == 'http':
            await self.handle_http_request(scope, receive, send)

        elif scope['type'] == 'websocket':
            await self.handle_websocket_request(scope, receive, send)

        else:
            await response_405(scope, receive, send)

    async def handle_lifespan(
        self, scope: Scope, receive: Receive, send: Send
    ) -> int | None:
        message = await receive()

        if message['type'] == 'lifespan.startup':
            await send({'type': 'lifespan.startup.complete'})
            return None

        elif message['type'] == 'lifespan.shutdown':
            await send({'type': 'lifespan.shutdown.complete'})
            return EXIT_CODE

        else:
            return None

    async def handle_http_request(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        path = self.static_files.get_path(scope)

        if path == '.':
            resp = await self.static_files.get_response(
                self.INDEX_HTML_PATH, scope
            )
            await resp(scope, receive, send)

        elif any([path.endswith(ext) for ext in self.ALLOWED_STATIC_FILES]):
            await self.static_files(scope, receive, send)

        else:
            await response_404(scope, receive, send)

    async def handle_websocket_request(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        socket = WebSocket(scope=scope, receive=receive, send=send)
        await socket.accept()

        with self.session_cls(socket) as session:
            await self.on_session_open()
            try:
                while True:
                    event = await session.listen_event()
                    await self.on_event(event=event)

            except (SessionClosed, asyncio.CancelledError):
                pass

            except Exception as e:
                logger.exception(e)

            finally:
                await self.on_session_close()



from __future__ import annotations

import abc
import json
import logging
import typing as t
from contextvars import ContextVar

from starlette import websockets

from .messages import Command
from .messages import Event

logger = logging.getLogger(__name__)

_session: ContextVar[AbstractSession] = ContextVar('_session')


class SessionClosed(Exception): ...


class AbstractSession(abc.ABC):
    type ID = str
    id: ID

    def __enter__(self):
        _session.set(self)
        return self

    def __exit__(self, *_):
        _session.set(None)

    @classmethod
    def get(cls) -> t.Self:
        return _session.get()

    @abc.abstractmethod
    async def _listen_event(self) -> Event: ...

    @abc.abstractmethod
    async def _send_command(self, cmd: Command) -> None: ...

    async def listen_event(self) -> Event:
        event = await self._listen_event()

        logger.info(f'[{self.id}] EVN >> {event._name:>20}  {event._data}')
        return event

    async def send_command(self, cmd: Command) -> None:
        await self._send_command(cmd)

        fmt_data = cmd._data
        if 'html' in fmt_data:
            fmt_data['html'] = '...'  # TODO: Need to move to logging level

        logger.info(f'[{self.id}] CMD << {cmd._name:>20}  {fmt_data}')


class Session(AbstractSession):
    __id = 0

    @classmethod
    def new_id(cls) -> int:
        cls.__id += 1
        return cls.__id

    def __init__(self, socket: websockets.WebSocket):
        self.id = str(self.__class__.new_id())
        self.socket = socket

    def _parse_event(self, message: str) -> Event:
        name, data = message.split(" ", 1)
        event_cls = Event.get_by_name(name)

        return event_cls(**json.loads(data))

    async def _listen_event(self) -> Event:
        try:
            message = await self.socket.receive_text()
            return self._parse_event(message)

        except websockets.WebSocketDisconnect:
            raise SessionClosed

    async def _send_command(self, cmd: Command):
        cmd_name = cmd.__class__.__name__
        cmd_params = json.dumps(cmd.__dict__)
        await self.socket.send_text(f'{cmd_name} {cmd_params}')
